/************************************************************************/
/* pfm.h - Packet Filter Module	header					*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#ifndef _PFM_H_
#define _PFM_H_

#include <stdint.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <pnetbuffer.h>
#include <cutil/slist.h>

#include <ioloop.h>


#define PFM_VERSION_1		0x0001

#define PFM_HASH_SIZE		65536


#define PFM_COMMAND_STR_SIZE	64
#define PFM_COMMAND_TOKEN_MAX	64


typedef enum {
	PFM_OPT_NOOP			= 0,
	PFM_OPT_SET_LOG_CALLBACK,
} pfm_option_t;


typedef enum {
	PFM_ERR_SUCCESS			= 0,
	PFM_ERR_INVALID_MOD,
	PFM_ERR_MOD_VERSION,
	PFM_ERR_CMD_INVALID,
	PFM_ERR_MOD_INTERNAL,
} pfm_errno_t;


typedef enum {
	PFM_IOCTL_BASE			= 0,	// noop
	PFM_IOCTL_REGISTER_BACKEND_FD,		// register a backend FD with
						// the I/O loop
	PFM_IOCTL_UNREGISTER_BACKEND_FD,	// unregister a backend FD with
						// the I/O loop
	_PFM_IOCTL_MAXIMUM,
} pfm_ioctl_option_t;


// argument to PFM_IOCTL_REGISTER_BACKEND_FD ioctl calls
typedef struct _ioctl_opt_register_fd_t ioctl_opt_register_fd_t;
struct _ioctl_opt_register_fd_t {
	int			fd;
	uint16_t		port;
	ioloop_callback_func	callback;
	void			*callback_arg;
};


typedef struct _io_context_t io_context_t;
struct _io_context_t {
	int			src_fd;
	struct sockaddr		src_addr;
	uint16_t		dest_port;
};


typedef struct _pfm_t pfm_t;
struct _pfm_t {
	uint16_t	version;
	uint16_t	flags;
	char		*name;
	uint16_t	*port_binds;
	void		*vectors; // use .version to determine structure version
	void		*userdata; // module private data pointer
};


// vector table filled by the module in it's pfm_module_open() call
typedef struct _pfm_vectors_t pfm_vectors_t;
struct _pfm_vectors_t {
	int	(*process_command)(pfm_t *pfm,
				pnetbuffer_t *indata,
				pnetbuffer_t *outdata);
	int	(*process_data)(pfm_t *pfm,
				io_context_t *ctx,
				pnetbuffer_t *indata,
				pnetbuffer_t *outdata);
};



typedef struct _pfm_module_t pfm_module_t;

typedef int (*pfm_module_info_func)(pfm_t *pfm);
typedef int (*pfm_module_open_func)(pfm_module_t *module);
typedef int (*pfm_module_close_func)(void *args);
typedef int (*pfm_log_func)(int log_level, char *format, va_list args);

struct _pfm_module_t {
	void			*dlmodule;
	void			*manager;	// opaque manager pointer
	pfm_t			*info;
	pfm_module_open_func	pfm_module_open;
	pfm_module_close_func	pfm_module_close;
};


typedef struct _pfm_init_t pfm_init_t;
struct _pfm_init_t {
	ioloop_t	*ioloop;
};


typedef struct _pfm_manager_t pfm_manager_t;
struct _pfm_manager_t {
	pfm_init_t		*provider;
	pfm_log_func		log_callback;

	SList			*module_list;

	// hash of SLists for port/module bindings
	SList			*port_hash[PFM_HASH_SIZE];
};




int pfm_pnetbuffer_add_rc_string(pnetbuffer_t *out, uint16_t err);

pfm_manager_t *pfm_open(pfm_init_t *init);
int pfm_set_option(pfm_manager_t *pfm, pfm_option_t option, void *arg);
int pfm_close(pfm_manager_t *pfm);
SList *pfm_module_list(pfm_manager_t *pfm);
int pfm_process_packet(pfm_manager_t *manager,
			io_context_t *ctx,
			pnetbuffer_t *in,
			pnetbuffer_t *out);
int pfm_process_command(pfm_manager_t *manager,
			pnetbuffer_t *in,
			pnetbuffer_t *out);

int pfm_module_load(pfm_manager_t *manager, char *path);
int pfm_module_unload(pfm_manager_t *manager, char *name);
int pfm_module_bind_module_to_port(pfm_manager_t *manager,
				char *name,
				uint16_t port);


// vector functions for modules to use to communicate back to PFM subsystem
int pfm_vector_ioctl(pfm_module_t *module, int option, void *optionarg);
int pfm_vector_log(pfm_module_t *module, int log_level, char *format, ...);


#endif


