/************************************************************************/
/* pfm_dns.c - DNS packet filter module					*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cutil/cutil.h>
#include <pcre.h>

#include <pnetbuffer.h>
#include <logging.h>
#include <udp.h>
#include <pfm.h>



#define PACKET_BUFFER_SIZE	16384
#define PFM_DNS_PORT_DEFAULT	53
#define RESPONSE_ARRAY_SIZE	4
#define RESPONSE_STRING_SIZE	128


#define VOTE_PASS			0
#define VOTE_DENY			1
#define VOTE_DROP			2
#define VOTE_LOG			3


#define DNS_F_QR			0x8000
#define DNS_F_AA			0x0400
#define DNS_F_TC			0x0200
#define DNS_F_RD			0x0100
#define DNS_F_RA			0x0080

#define DNS_F_OP_0			0x0080
#define DNS_F_OP_1			0x0100
#define DNS_F_OP_2			0x0200
#define DNS_F_OP_3			0x0400

#define DNS_F_Z_0			0x0010
#define DNS_F_Z_1			0x0020
#define DNS_F_Z_2			0x0040

#define DNS_F_RC_0			0x0001
#define DNS_F_RC_1			0x0002
#define DNS_F_RC_2			0x0004
#define DNS_F_RC_3			0x0008
#define DNS_F_RC_SUCCESS		0
#define DNS_F_RC_FMTERROR		1
#define DNS_F_RC_SRVERROR		2
#define DNS_F_RC_NXDOMAIN		3
#define DNS_F_RC_NOIMPL			4
#define DNS_F_RC_REFUSED		5


#define HASH_TABLE_SIZE_DEFAULT		2048

const char *vote_string[] = {
	"pass",
	"deny",
	"drop",
	"log",
};


#define F_TYPE_SUBSTRING		1
#define F_TYPE_REGEX			2

typedef struct _filter_t filter_t;
struct _filter_t {
	uint8_t		type;
	uint8_t		vote;
	char		*match;
	void		*arg;
};


typedef struct _counters_t counters_t;
struct _counters_t {
	uint32_t	q_total;
	uint32_t	q_filtered;
	uint32_t	an_total;
	uint32_t	an_filtered;
};


typedef struct _request_t request_t;
struct _request_t {
	// Source identification
	union {
		uint64_t	hash;
		struct {
			uint32_t	s_ip;
			uint16_t	s_port;
			uint16_t	req_id;
		} part;
	} u;
	struct sockaddr	src_sockaddr;

	// frontend/backend tracking
	int		frontend_fd;
	int		backend_id; // table index
};


#define DNS_BACKEND_CONNECTIONS_DEFAULT		4
typedef struct _dns_backend_t dns_backend_t;
struct _dns_backend_t {
	char		*dest_host;
	char		*dest_service;
	struct sockaddr	dest_sa;

	int		conn_count;
	int		conn_fd[DNS_BACKEND_CONNECTIONS_DEFAULT];
};


#define DNS_BACKENDS_MAX		4
typedef struct _pfm_dns_t pfm_dns_t;
struct _pfm_dns_t {
	pfm_module_t	*module;	// passed in module_open()

	uint64_t	policy_bits;

	char		**command_array;
	char		**response_array;

	counters_t	counters;

	SList		*q_filter_list;
	SList		*an_filter_list;

	int		req_table_size;
	request_t	*req_table;

	int		backend_count;			// configured backends
	dns_backend_t	backend[DNS_BACKENDS_MAX];	// array of backends
};


// policy bits management
typedef enum {
	POLICY_NOOP				= 0,
	POLICY_DENY_MULTIPLE_QUERIES,

	_POLICY_MAXIMUM,
} policy_bit_t;

typedef struct _policy_item_t policy_item_t;
struct _policy_item_t {
	char		*name;
	policy_bit_t	bit; // policy_bits |= (1<<bit)
};

policy_item_t policy_table[] = {
	{"deny-multiquery",	POLICY_DENY_MULTIPLE_QUERIES},
	{NULL,			POLICY_NOOP},
};



// Forward prototypes
int vector_process_command(pfm_t *pfm, pnetbuffer_t *indata, pnetbuffer_t *outdata);
int vector_process_data(pfm_t *pfm, io_context_t *ctx, pnetbuffer_t *indata, pnetbuffer_t *outdata);
int backend_io_callback(int fd, int events, connection_data_t *cd, void *arg);


// Module Globals
pfm_vectors_t vectors = {
	.process_command	= vector_process_command,
	.process_data		= vector_process_data,
};

uint16_t port_list[] = {
	PFM_DNS_PORT_DEFAULT,
	0,
};

pfm_t pfm_info = {
	.version	= PFM_VERSION_1,
	.flags		= 0,
	.name		= "dns",
	.vectors	= &vectors,
	.userdata	= NULL,
};


pfm_dns_t	priv_data;



int backend_increase_connections(dns_backend_t *be, int count)
{
	int			rc;
	int			fd;
	int			i;
	ioctl_opt_register_fd_t	opt;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"Connection increase requested for backend %p count=%d",
		be,
		count);

	for (i=0 ; i<count ; i++) {
		fd = udp_connect(be->dest_host, be->dest_service);
		if (fd == -1)
			return errno;
		opt.fd = fd;
		opt.port = 0;
		opt.callback = backend_io_callback;
		opt.callback_arg = NULL;
		rc = pfm_vector_ioctl(priv_data.module,
				PFM_IOCTL_REGISTER_BACKEND_FD,
				&opt);
		if (rc != 0)
			return rc;
		be->conn_fd[be->conn_count] = fd;
		be->conn_count++;
	}
	return 0;
}


int backend_decrease_connections(dns_backend_t *be, int count)
{
	int			rc;
	int			i;
	int			last;
	ioctl_opt_register_fd_t	opt;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"Connection decrease requested for backend %p count=%d",
		be,
		count);

	if (count > be->conn_count)
		return E2BIG;
	for (i=0 ; i<count ; i++) {
		last = be->conn_count-1;
		opt.fd = be->conn_fd[last];
		opt.port = 0;
		opt.callback = backend_io_callback;
		opt.callback_arg = NULL;
		rc = pfm_vector_ioctl(priv_data.module,
				PFM_IOCTL_UNREGISTER_BACKEND_FD,
				&opt);
		if (rc != 0) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"Failed to close connection for backend %p fd=%d rc=%d",
				be,
				be->conn_fd[last],
				rc);
			return rc;
		}
		rc = udp_shutdown(be->conn_fd[last]);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"Connection shutdown() for backend %p fd=%d rc=%d",
			be,
			be->conn_fd[last],
			rc);
		be->conn_fd[last] = -1;
		be->conn_count--;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"Connection closed for backend %p oset=%d",
			be,
			last);
	}
	return 0;
}


int backend_replace_connection(dns_backend_t *be, int fd)
{
	int			rc;
	int			i;
	ioctl_opt_register_fd_t	opt;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"Replacement connection requested for backend %s:%s fd=%d",
		be->dest_host,
		be->dest_service,
		fd);

	// Find old fd in table
	for (i=0 ; i<be->conn_count ; i++) {
		if (be->conn_fd[i] == fd)
			break;
	}
	if (i == be->conn_count)
		return ENOENT;

	// Close the old connection
	opt.fd = be->conn_fd[i];
	opt.port = 0;
	opt.callback = backend_io_callback;
	opt.callback_arg = NULL;
	rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_UNREGISTER_BACKEND_FD,
			&opt);
	if (rc != 0)
		return rc;
	(void)udp_shutdown(be->conn_fd[i]);
	be->conn_fd[i] = -1;

	// Create the replacement connection
	fd = udp_connect(be->dest_host, be->dest_service);
	if (fd == -1)
		return errno;
	opt.fd = fd;
	opt.port = 0;
	opt.callback = backend_io_callback;
	opt.callback_arg = NULL;
	rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_REGISTER_BACKEND_FD,
			&opt);
	if (rc != 0)
		return rc;
	be->conn_fd[i] = fd;

	return 0;
}


int backend_add(char *hostname, char *service, int connections)
{
	int			rc;
	dns_backend_t		*be;
	struct addrinfo		hint;
	struct sockaddr_in	*sa_in;
	struct addrinfo		*res;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"New backend requested for %s:%s with %d connections",
		hostname,
		service,
		connections);

	be = &priv_data.backend[priv_data.backend_count];
	be->dest_host = strdup(hostname);
	be->dest_service = strdup(service);

	memset(&hint, 0, sizeof(struct addrinfo));
	hint.ai_family = AF_INET;
	hint.ai_socktype = SOCK_DGRAM;
	hint.ai_protocol = IPPROTO_UDP;
	rc = getaddrinfo(be->dest_host,
			be->dest_service,
			&hint,
			&res);
	if (rc != 0) {
		free(be->dest_host);
		free(be->dest_service);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"%s()  getaddrinfo() failed rc=%d\n",
			__func__,
			rc);
		return rc;
	}
	sa_in = (struct sockaddr_in *)res->ai_addr;
	memcpy(&be->dest_sa, sa_in, sizeof(struct sockaddr_in));
	freeaddrinfo(res);

	rc = backend_increase_connections(be, connections);
	if (rc != 0)
		return rc;
	priv_data.backend_count++;
	return 0;
}


int backend_remove(char *hostname, char *service)
{
	int			rc;
	int			i;
	int			be_oset;
	int			be_last;
	char			*h;
	char			*s;
	dns_backend_t		*be = NULL;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"Remove backend requested for %s:%s",
		hostname,
		service);

	// Find old backend in table
	for (i=0 ; i<priv_data.backend_count ; i++) {
		be = &priv_data.backend[i];
		if (!strcmp(be->dest_host, hostname) &&
			!strcmp(be->dest_service, service))
			break;
	}
	if (i == priv_data.backend_count)
		return ENOENT;

	// get references
	be_oset = i;
	h = be->dest_host;
	s = be->dest_service;

	// drain connections
	rc = backend_decrease_connections(be, be->conn_count);
	if (rc) return rc;

	// remove backend from table (slide down)
	be_last = priv_data.backend_count-1;
	for (i=be_oset ; i<be_last ; i++) {
		memcpy(&priv_data.backend[i],
			&priv_data.backend[i+1],
			sizeof(dns_backend_t));
	}
	memset(&priv_data.backend[be_last], 0, sizeof(dns_backend_t));
	priv_data.backend_count--;

	// free resources
	free(h);
	free(s);

	return 0;
}



int pfm_module_info(pfm_t *pfm)
{
	pfm->version = pfm_info.version;
	pfm->flags = pfm_info.flags;
	pfm->name = pfm_info.name;
	pfm->vectors = pfm_info.vectors;
	pfm->userdata = pfm_info.userdata;
	return 0;
}


int pfm_module_open(pfm_module_t *module)
{
	int		rc;
	int		i;

	memset(&priv_data, 0, sizeof(pfm_dns_t));
	priv_data.module = (pfm_module_t *)module;
	// assign default port(s)
	pfm_info.port_binds = port_list;

	priv_data.req_table_size = HASH_TABLE_SIZE_DEFAULT;
	priv_data.req_table = calloc(priv_data.req_table_size, sizeof(request_t));
	if (!priv_data.req_table) {
		rc = errno;
		goto pfm_module_open_exit;
	}

	// allocate command buffer array
	priv_data.command_array = calloc(PFM_COMMAND_TOKEN_MAX, sizeof(char*));
	if (!priv_data.command_array) {
		rc = errno;
		goto pfm_module_open_exit;
	}
	for (i=0 ; i<PFM_COMMAND_TOKEN_MAX ; i++) {
		priv_data.command_array[i] = calloc(1, PFM_COMMAND_STR_SIZE);
		if (!priv_data.command_array[i]) {
			rc = errno;
			goto pfm_module_open_exit;
		}
	}

	// allocate response buffer array
	priv_data.response_array = calloc(RESPONSE_ARRAY_SIZE, sizeof(char*));
	if (!priv_data.response_array) {
		rc = errno;
		goto pfm_module_open_exit;
	}
	for (i=0 ; i<RESPONSE_ARRAY_SIZE ; i++) {
		priv_data.response_array[i] = calloc(1, RESPONSE_STRING_SIZE);
		if (!priv_data.response_array[i]) {
			rc = errno;
			goto pfm_module_open_exit;
		}
	}
	return 0;

pfm_module_open_exit:
	if (priv_data.command_array)
		free(priv_data.command_array);
	if (priv_data.response_array)
		free(priv_data.response_array);
	return rc;
}


int pfm_module_close(void *args)
{
	if (priv_data.command_array)
		free(priv_data.command_array);
	return 0;
}



policy_item_t *policy_item_lookup_by_name(char *name)
{
	int		i = 0;
	while (policy_table[i].name != NULL) {
		if (!strcmp(name, policy_table[i].name))
			return &policy_table[i];
		i++;
	}
	return NULL;
}


int vector_process_command(pfm_t *pfm, pnetbuffer_t *indata, pnetbuffer_t *outdata)
{
	int		rc;
	int		i;
	int		count;
	uint8_t		argc;
	char		**argv;
	char		*str1;
	char		*str2;
	filter_t	*filter;
	SList		*li;
	SList		**lp;
	policy_item_t	*politem;
	
	rc = pnetbuffer_read_strings(indata,
				priv_data.command_array,
				PFM_COMMAND_STR_SIZE,
				PFM_COMMAND_TOKEN_MAX,
				&argc,
				PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
	if (rc != 0) {
		(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
		return rc;
	}
	if (argc < 1) {
		(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
		return 0;
	}
	argv = priv_data.command_array;
	i = 0;

	if (!strcmp(argv[i], "counters")) {
		i++;
		if (!strcmp(argv[i], "show")) {
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
			snprintf (priv_data.response_array[0],
				128,
				"q_total=%d q_filtered=%d an_total=%d an_filtered=%d\n",
				priv_data.counters.q_total,
				priv_data.counters.q_filtered,
				priv_data.counters.an_total,
				priv_data.counters.an_filtered);
			rc = pnetbuffer_add_strings(outdata,
						priv_data.response_array,
						1,
						PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
			if (rc != 0) {
				pnetbuffer_reset(outdata);
				(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_MOD_INTERNAL);
				return 0;
			}
		} else if (!strcmp(argv[i], "reset")) {
			memset(&priv_data.counters, 0, sizeof(counters_t));
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
		} else {
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
			return 0;
		}
		return 0;

	} else if (!strcmp(argv[i], "backend")) {
		i++;
		if (!strcmp(argv[i], "add")) {
			i++;
			if (!strcmp(argv[i], "host")) {
				i++;
				str1 = argv[i];
				i++;
				if (!strcmp(argv[i], "port")) {
					i++;
					str2 = argv[i];
					i++;
					if (!strcmp(argv[i], "conns")) {
						i++;
						count = atoi(argv[i]);
						rc = backend_add(str1, str2, count);
						if (rc != 0) {
							(void)pfm_pnetbuffer_add_rc_string(outdata,
										PFM_ERR_CMD_INVALID);
							return 0;
						}
						(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
						return 0;
					} else {
						(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
						return 0;
					}
				} else {
					(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
					return 0;
				}
			} else {
				(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
				return 0;
			}
		} else if (!strcmp(argv[i], "remove")) {
			i++;
			if (!strcmp(argv[i], "host")) {
				i++;
				str1 = argv[i];
				i++;
				if (!strcmp(argv[i], "port")) {
					i++;
					str2 = argv[i];
					// TODO: remove the backend
					rc = backend_remove(str1, str2);
					if (rc != 0) {
						(void)pfm_pnetbuffer_add_rc_string(outdata,
									PFM_ERR_CMD_INVALID);
						return 0;
					}
				} else {
					(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
					return 0;
				}
			} else {
				(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
				return 0;
			}
		} else {
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
			return 0;
		}
	} else if (!strcmp(argv[i], "policy")) {
		i++;
		lp = NULL;
		if (!strcmp(argv[i], "list")) {
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
			i = 0;
			while (policy_table[i].name != NULL) {
				if (priv_data.policy_bits & (1<<policy_table[i].bit)) {
					snprintf (priv_data.response_array[0],
						128,
						"policy set %s\n",
						policy_table[i].name);
					rc = pnetbuffer_add_strings(outdata,
								priv_data.response_array,
								1,
								PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
					if (rc != 0) {
						pnetbuffer_reset(outdata);
						(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_MOD_INTERNAL);
						return 0;
					}
				}
				i++;
			}
			return 0;
		} else if (!strcmp(argv[i], "set")) {
			i++;
			politem = policy_item_lookup_by_name(argv[i]);
			if (!politem) {
				(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
				return 0;
			}
			priv_data.policy_bits |= (1<<politem->bit);
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
			return 0;
		} else if (!strcmp(argv[i], "unset")) {
			i++;
			politem = policy_item_lookup_by_name(argv[i]);
			if (!politem) {
				(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
				return 0;
			}
			priv_data.policy_bits &= ~(1<<politem->bit);
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
			return 0;
		} else {
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
			return 0;
		}

	} else if (!strcmp(argv[i], "filter")) {
		i++;
		lp = NULL;
		if (!strcmp(argv[i], "list")) {
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);

			li = priv_data.q_filter_list;
			while (li) {
				filter = slist_get_data(li);
				snprintf (priv_data.response_array[0],
					128,
					"filter query \"%s\" %s\n",
					filter->match, vote_string[filter->vote]);
				rc = pnetbuffer_add_strings(outdata,
							priv_data.response_array,
							1,
							PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
				if (rc != 0) {
					pnetbuffer_reset(outdata);
					(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_MOD_INTERNAL);
					return 0;
				}
				li = slist_get_next(li);
			}
			return 0;
		} else if (!strcmp(argv[i], "add")) {
			filter = calloc(1, sizeof(filter_t));
			if (!filter) {
				pnetbuffer_reset(outdata);
				(void)pfm_pnetbuffer_add_rc_string(outdata,
							PFM_ERR_MOD_INTERNAL);
				return 0;
			}
			// element to filter (query/answer)
			i++;
			if (!strcmp(argv[i], "query")) {
				lp = &priv_data.q_filter_list;
			} else if (!strcmp(argv[i], "answer")) {
				lp = &priv_data.an_filter_list;
			} else {
				pnetbuffer_reset(outdata);
				(void)pfm_pnetbuffer_add_rc_string(outdata,
							PFM_ERR_CMD_INVALID);
				return 0;
			}
			// filter match string
			i++;
			if (argv[i][0] == '/') {
				//TODO compile regex and store into filter->arg
				count = strlen(argv[i]);
				if (count >= 3 && argv[i][count-1] == '/') {
					filter->type = F_TYPE_REGEX;
					argv[i][count-1] = 0;
					filter->match = strdup(&argv[i][1]);
					filter->arg = pcre_compile(filter->match,
								0,
								(const char **)&str1,
								&count,
								NULL);
					if (!filter->arg) {
						pnetbuffer_reset(outdata);
						(void)pfm_pnetbuffer_add_rc_string(outdata,
									PFM_ERR_CMD_INVALID);
						return 0;
					}
				} else {
					pnetbuffer_reset(outdata);
					(void)pfm_pnetbuffer_add_rc_string(outdata,
									PFM_ERR_CMD_INVALID);
					return 0;
				}
			} else {
				filter->type = F_TYPE_SUBSTRING;
				filter->match = strdup(argv[i]);
			}
			// match vote
			i++;
			if (!strcmp(argv[i], "allow")) {
				filter->vote = VOTE_PASS;
			} else if (!strcmp(argv[i], "deny")) {
				filter->vote = VOTE_DENY;
			} else if (!strcmp(argv[i], "log")) {
				filter->vote = VOTE_LOG;
			} else {
				pnetbuffer_reset(outdata);
				(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
				return 0;
			}
			// store filter
			rc = slist_item_add(lp, filter);
			if (rc) {
				pnetbuffer_reset(outdata);
				(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_MOD_INTERNAL);
				return 0;
			}
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
			return 0;
		} else if (!strcmp(argv[i], "remove")) {
		} else {
			(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
			return 0;
		}
		return 0;
	} else if (!strcmp(argv[i], "test")) {
		(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_SUCCESS);
		return 0;
	} else {
		(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
		return 0;
	}
	return 0;
}


int filter_match(filter_t * filter, char *str)
{
	int	rc;
	int	ovector[32];
	switch (filter->type) {
		case F_TYPE_SUBSTRING:
			if (strstr(str, filter->match))
				return 1;
			break;
		case F_TYPE_REGEX:
			// TODO regex match
			rc = pcre_exec(filter->arg,	// pcre
					NULL,		// extra
					str,		// subject str
					strlen(str),	// length
					0,		// start oset
					0,		// options
					ovector,	// ovect
					32);		// ovect size
			if (rc >= 1)
				return 1;
			break;
	}
	return 0;
}


//int request_table_find_free(void)
//{
//	int		i;
//	for (i=0 ; i<priv_data.req_table_size ; i++) {
//		if (!(priv_data.req_table[i].flags & TABLE_ITEM_INUSE)
//			return i;
//	}
//	return -1;
//}


int request_table_hash(int fd, uint16_t req_id)
{
	return ((((fd & 0xffff)<<16) + req_id) % priv_data.req_table_size);
}


int pnetbuffer_sendto(pnetbuffer_t *nb, int fd, int flags, struct sockaddr *sa, socklen_t salen)
{
	int		rc;
	uint16_t	plen;
	char		*p;

	pnetbuffer_get_readable(nb, &p, &plen);
	rc = sendto(fd, p, plen, flags, sa, salen);
	return rc;
}


int dns_packet_read_header(pnetbuffer_t *indata,
			uint16_t *id,
			uint16_t *flags,
			uint16_t *q_count,
			uint16_t *a_count,
			uint16_t *ns_count,
			uint16_t *ar_count)
{
	int		rc;
	rc = pnetbuffer_read_uint16(indata, id);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, flags);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, q_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, a_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, ns_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint16(indata, ar_count);
	if (rc != 0) return rc;
	return 0;
}


int dns_packet_write_header(pnetbuffer_t *nb,
			uint16_t id,
			uint16_t flags,
			uint16_t q_count,
			uint16_t a_count,
			uint16_t ns_count,
			uint16_t ar_count)
{
	int		rc;
//	printf("TRACE %s() id=0x%04x\n", __func__, id);
//	printf("TRACE %s() flags=0x%04x\n", __func__, flags);
//	printf("TRACE %s() q_count=0x%04x\n", __func__, q_count);
//	printf("TRACE %s() a_count=0x%04x\n", __func__, a_count);
//	printf("TRACE %s() ns_count=0x%04x\n", __func__, ns_count);
//	printf("TRACE %s() ar_count=0x%04x\n", __func__, ar_count);
	rc = pnetbuffer_add_uint16(nb, id);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, flags);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, q_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, a_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, ns_count);
	if (rc != 0) return rc;
	rc = pnetbuffer_add_uint16(nb, ar_count);
	if (rc != 0) return rc;
//	printf("TRACE %s() exit successfully\n", __func__);
	uint16_t plen;
	char	*p;
	pnetbuffer_get_readable(nb, &p, &plen);
//	printf("TRACE %s() p=%p len=%d\n", __func__, p, plen);
	return 0;
}


int send_response(int fd, struct sockaddr *dest, pnetbuffer_t *nb)
{
	int		rc;
	uint16_t	plen;
	char		*p;

	// forward response to source
	pnetbuffer_get_readable(nb, &p, &plen);
//	printf("TRACE %s() p=%p len=%d\n", __func__, p, plen);
	rc = sendto(fd,
			p,
			plen,
			0,
			dest,
			sizeof(struct sockaddr));
	if (rc == -1) {
		rc = errno;
		if (rc == EAGAIN || rc == EWOULDBLOCK) {
			// TODO Mark frontend socket to notify when writable again
			// TODO and we'll need to save this packet to try again later
		}
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"sendto(%d, %p, %d, %d, %p, %lu) rc=%d",
			fd,
			p,
			plen,
			0,
			dest,
			sizeof(struct sockaddr),
			rc);
		return rc;
	}
//	printf("TRACE %s() exit successfully\n", __func__);
	return 0;
}


int dns_fqdn_write_netbuffer(pnetbuffer_t *nb, char *str)
{
	int		rc;
	uint8_t		len;
	char		*p;
	char		*pp;
	p = str;
	while (1) {
		pp = strstr(p, ".");
		if (pp) {
			if (pp == p) {
				p++;
				continue;
			}
			len = pp - p;
		} else {
			len = strlen(p);
		}
		rc = pnetbuffer_add_uint8(nb, len);
		if (rc != 0) return rc;
		rc = pnetbuffer_add_data(nb, p, len);
		if (rc != 0) return rc;
//		printf("TRACE %s() p=%s len=%d\n", __func__, p, len);
		if (!pp) break; // end of string
		p = pp++;
	}

//	printf("TRACE %s() exit successfully\n", __func__);
	return 0;

}


int vector_process_data(pfm_t *pfm, io_context_t *ctx, pnetbuffer_t *indata, pnetbuffer_t *outdata)
{
	int			rc;
	int				i;
	int			req_id;
	int			conn_id;
	int			vote;
	uint16_t		u16;
	uint16_t		id;
	uint16_t		flags;
	uint16_t		q_count;
	uint16_t		a_count;
	uint16_t		ns_count;
	uint16_t		ar_count;
	uint16_t		type;
	uint16_t		class;
	uint8_t			len;
	char			str[1024];
	char			*p;
	SList			*item;
	filter_t		*filter;
	request_t		req;
	struct sockaddr_in	*sin;

	vote = VOTE_PASS;

	rc = dns_packet_read_header(indata,
				&id,
				&flags,
				&q_count,
				&a_count,
				&ns_count,
				&ar_count);
	if (rc != 0) return VOTE_DROP; //malformed

//	printf("DEBUG dns request id=0x%04x q_count=%d\n", id, q_count);

	if (priv_data.policy_bits & POLICY_DENY_MULTIPLE_QUERIES) {
		sin = (struct sockaddr_in *)&ctx->src_addr;
		printf("WARN dns request denied by policy: multiple queries present source=%s q_count=%d type=%d class=%d\n",
			inet_ntoa(sin->sin_addr),
			q_count,
			type,
			class);
		priv_data.counters.q_filtered++;
		return VOTE_DENY;
	}

	// process query section
	for (i=0 ; i<q_count ; i++) {
		p = str;
		*p = 0;
		while (1) {
			rc = pnetbuffer_read_uint8(indata, &len);
			if (rc != 0) return VOTE_DROP; //malformed
			if (len == 0) break;
			rc = pnetbuffer_read_data(indata, p, len);
			if (rc != 0) return VOTE_DROP; //malformed
			p += len;
			*p = '.';
			p++;
		}
		*p = 0;
		rc = pnetbuffer_read_uint16(indata, &type);
		if (rc != 0) return VOTE_DROP; //malformed
		rc = pnetbuffer_read_uint16(indata, &class);
		if (rc != 0) return VOTE_DROP; //malformed
//		printf("DEBUG dns query name='%s' type=%d class=%d\n", str, type, class);
		priv_data.counters.q_total++;

		// filter query
		item = priv_data.q_filter_list;
		while (item) {
			filter = slist_get_data(item);
//			printf("DEBUG filter->match='%s' name='%s' \n", filter->match, str);
			if (filter_match(filter, str)) {
				switch (filter->vote) {
					case VOTE_DENY:
						sin = (struct sockaddr_in *)&ctx->src_addr;
						pfm_vector_log(priv_data.module,
							LOG_LEVEL_WARNING,
							"dns query denied source=%s name='%s' type=%d class=%d",
							inet_ntoa(sin->sin_addr),
							str,
							type,
							class);
						priv_data.counters.q_filtered++;

						// send NXDOMAIN response
						u16 = DNS_F_QR|DNS_F_AA|DNS_F_RC_NXDOMAIN;
						if (flags & DNS_F_RD)
							u16 |= (DNS_F_RD|DNS_F_RA);
						pnetbuffer_reset(outdata);
						rc = dns_packet_write_header(outdata,
									id,
									u16,
									q_count,
									0,
									0,
									0);
						if (rc != 0) return VOTE_DENY; //buffer size
						rc = dns_fqdn_write_netbuffer(outdata, str);
						if (rc != 0) return VOTE_DENY; //buffer size
						rc = pnetbuffer_add_uint16(outdata, type);
						if (rc != 0) return VOTE_DENY; //buffer size
						rc = pnetbuffer_add_uint16(outdata, class);
						if (rc != 0) return VOTE_DENY; //buffer size
						rc = send_response(ctx->src_fd, &ctx->src_addr, outdata);
						if (rc != 0) {
							pfm_vector_log(priv_data.module,
								LOG_LEVEL_WARNING,
								"Failed to send nxdomain response source=%s name='%s'",
								inet_ntoa(sin->sin_addr),
								str);
						}
						return VOTE_DENY; //io error
						break;
					case VOTE_LOG:
						sin = (struct sockaddr_in *)&ctx->src_addr;
						pfm_vector_log(priv_data.module,
							LOG_LEVEL_WARNING,
							"dns query filter match source=%s name='%s' type=%d class=%d",
							inet_ntoa(sin->sin_addr),
							str,
							type,
							class);
						break;
				}
			}
			item = slist_get_next(item);
		}

	}

	if (priv_data.backend_count == 0) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"No backend configured, dropping packet");
		return VOTE_DENY; //config error
	}

	// select backend and connection to use
	req.backend_id = (random() % priv_data.backend_count);
	conn_id = (random() % priv_data.backend[req.backend_id].conn_count);


	req_id = request_table_hash(priv_data.backend[req.backend_id].conn_fd[conn_id],
					id);
//	printf("DEBUG %s() using req_table[%d]\n", __func__, req_id);

	// setup request table entry
	req.src_sockaddr = ctx->src_addr;
	req.u.part.s_ip = ((struct sockaddr_in *)ctx)->sin_addr.s_addr;
	req.u.part.s_port = ((struct sockaddr_in *)ctx)->sin_port;
	req.u.part.req_id = id;
	req.frontend_fd = ctx->src_fd;
	priv_data.req_table[req_id] = req;

	// forward request to backend for processing
	pnetbuffer_reset_read(indata);
	pnetbuffer_get_readable(indata, &p, &u16);
	rc = sendto(priv_data.backend[req.backend_id].conn_fd[i],
			p,
			u16,
			0,
			(struct sockaddr *)&priv_data.backend[req.backend_id].dest_sa,
			sizeof(struct sockaddr));
	if (rc == -1) {
		rc = errno;
		if (rc == EAGAIN || rc == EWOULDBLOCK) {
			// TODO Mark backend socket to notify when writable again
			// TODO and we'll need to save this packet to try again later
		}
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"sendto() rc=%d backend=%s:%s\n",
			rc,
			priv_data.backend[req.backend_id].dest_host,
			priv_data.backend[req.backend_id].dest_service);
		return vote; //io error
	}

	return vote; //success
}



int backend_io_callback(int fd, int events, connection_data_t *cd, void *arg)
{
	int			rc;
	int			req_id;
	uint16_t		plen;
	uint16_t		id;
	uint16_t		flags;
	uint16_t		q_count;
	uint16_t		a_count;
	uint16_t		ns_count;
	uint16_t		ar_count;
	int			read_len;
	int			vote;
	socklen_t		saddr_len;
	char			buffer_in[PACKET_BUFFER_SIZE];
	char			*p;
	pnetbuffer_t		*nb_in;
	io_context_t		ctx;

	//************************************
	// init
	//************************************
	vote =  VOTE_PASS;
	ctx.dest_port = cd->port;

	//************************************
	// read packet
	//************************************
	nb_in = pnetbuffer_init(buffer_in, PACKET_BUFFER_SIZE);
	pnetbuffer_get_writable(nb_in, &p, &plen);
	saddr_len = sizeof(struct sockaddr);
	read_len = recvfrom(fd, p, (int)plen, 0, &ctx.src_addr, &saddr_len);
	if (read_len == -1) {
		rc = errno;
		return rc;
	}
	rc = pnetbuffer_set_written(nb_in, (uint16_t)read_len);
//	printf("DEBUG %s() response packet len=%d\n", __func__, read_len);

	rc = dns_packet_read_header(nb_in,
				&id,
				&flags,
				&q_count,
				&a_count,
				&ns_count,
				&ar_count);
	if (rc != 0) return VOTE_DROP;
	priv_data.counters.an_total++;
	
	req_id = request_table_hash(fd, id);
//	printf("DEBUG %s() using req_table[%d]\n", __func__, req_id);

	// forward response to source
	pnetbuffer_reset_read(nb_in);
	pnetbuffer_get_readable(nb_in, &p, &plen);
	rc = sendto(priv_data.req_table[req_id].frontend_fd,
			p,
			plen,
			0,
			&priv_data.req_table[req_id].src_sockaddr,
			sizeof(struct sockaddr));
	if (rc == -1) {
		rc = errno;
		if (rc == EAGAIN || rc == EWOULDBLOCK) {
			// TODO Mark backend socket to notify when writable again
			// TODO and we'll need to save this packet to try again later
		}
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"sendto(%d, %p, %d, %d, %p, %lu) rc=%d",
			priv_data.req_table[req_id].frontend_fd,
			p,
			plen,
			0,
			&priv_data.req_table[req_id].src_sockaddr,
			sizeof(struct sockaddr),
			rc);
		return vote;
	}

	return 0;
}


