/************************************************************************/
/* udpfilterctl.c							*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <signal.h>

#include <udp.h>
#include <helpers.h>
#include <logging.h>
#include <pnetbuffer.h>
#include <protocols.h>



#define PACKET_BUFFER_SIZE		16384


// Globals
logging_t *logger;


void usage(void)
{
	printf("Usage: udpfilterctl [options] command [arguments]\n");
	//printf("Options:\n");
	//printf("  -c cmdfile  Import a command decriptor file (one or more)\n");
	printf("Commands:\n");
	printf("  ping                             Send a ping packet to the daemon and await a response\n");
//	printf("  stats                            Request processing statistics from the daemon\n");
	printf("  pfm list                         Request list of loaded pfms from the daemon\n");
	printf("  pfm command pfmname args[...]    Send command to the specified pfm\n");
//	printf("  filter add|del allow|deny protocol [src N/N|dest N/N|sport N|dport N]\n");
//	printf("             Alter filter rules\n");
//	printf("  filter list protocol\n");
//	printf("             List filter rules for the protocol specified\n");
	printf("\n");
	return;
}



int command_and_response(pnetbuffer_t *cmd_nb, pnetbuffer_t *res_nb)
{
	int		rc;
	int		fd;
	uint16_t	len;
	uint16_t	rlen;
	nfds_t		nfds;
	struct pollfd	fds[2];
	char		*p;

	// connect to daemon
	fd = udp_connect("127.0.0.1", "5200");
	if (fd == -1) {
		logging_entry(logger,
			LOG_LEVEL_ERROR,
			"Failed to connect to udpfilterd command socket");
		return 2;
	}

	// send command
	pnetbuffer_get_readable(cmd_nb, &p, &len);
	rlen = send(fd, p, len, 0);
	if (rlen == -1) {
		logging_entry(logger,
			LOG_LEVEL_ERROR,
			"Failed to send command to udpfilterd errno=%d",
			errno);
		udp_shutdown(fd);
		return 2;
	}

	// await response
	nfds = 1;
	fds[0].fd = fd;
	fds[0].events = POLLIN;
	rc = poll(fds, nfds, 5000);
	if (rc == -1) {
		logging_entry(logger,
			LOG_LEVEL_ERROR,
			"Failed waiting on udpfilterd command response errno=%d",
			errno);
		udp_shutdown(fd);
		return 2;
	}
	if (rc == 0) { 
		logging_entry(logger,
			LOG_LEVEL_ERROR,
			"Timeout waiting on udpfilterd command response");
		udp_shutdown(fd);
		return 2;
	}

	// read/process response
	pnetbuffer_reset(res_nb);
	pnetbuffer_get_writable(res_nb, &p, &len);
	if (len == 0) {
		logging_entry(logger,
			LOG_LEVEL_ERROR,
			"Failed getting writable pointer for output buffer");
		udp_shutdown(fd);
		return 2;
	}
	rlen = recv(fd, p, (int)len, 0);
	if (rlen == -1) {
		logging_entry(logger,
			LOG_LEVEL_ERROR,
			"Failed reading from udpfilterd respoonse errno=%d",
			errno);
		udp_shutdown(fd);
		return 2;
	}
	rc = pnetbuffer_set_written(res_nb, rlen);
	udp_shutdown(fd);
	return rc;
}


int args_parse_compare_values(compare_t *c, char **args, int argc)
{
	int		i;
	uint16_t	compare;
	char		*p;

	c->compare = 0;
	compare = 0;
	i = 0;
	while (i < argc) {
		if (!strcmp(args[i], "proto")) {
			i++;
			c->protocol = protocol_to_l4_type(args[i]);
		} else if (!strcmp(args[i], "src")) {
			i++;
			// net
			p = strtok(args[i], "/");
			c->src_addr.net = atoi(p);
			// cidr->mask
			p = strtok(NULL, " ");
			c->src_addr.mask = ipv4_cidr_to_mask(atoi(p));
			compare |= COMPARE_ADDR_SRC;
		} else if (!strcmp(args[i], "dest")) {
			i++;
			// net
			p = strtok(args[i], "/");
			c->dest_addr.net = atoi(p);
			// cidr->mask
			p = strtok(NULL, " ");
			c->dest_addr.mask = ipv4_cidr_to_mask(atoi(p));
			compare |= COMPARE_ADDR_DEST;
		} else if (!strcmp(args[i], "sport")) {
			i++;
			c->dest_addr.net = (uint16_t)atoi(args[i]);
			compare |= COMPARE_PORT_SRC;
		} else if (!strcmp(args[i], "dport")) {
			i++;
			c->dest_addr.mask = (uint16_t)atoi(args[i]);
			compare |= COMPARE_PORT_DEST;
		} else if (!strcmp(args[i], "allow")) {
			c->action = COMPARE_ACTION_ALLOW;
		} else if (!strcmp(args[i], "deny")) {
			c->action = COMPARE_ACTION_DENY;
		} else {
			logging_entry(logger,
				LOG_LEVEL_ERROR,
				"invalid filter argument '%s'",
				args[i]);
			return EINVAL;
		}
		i++;
	}
	c->compare = compare;
	return 0;
}


int arguments_handle_pfm(pnetbuffer_t* cmd_nb, int argc, char **argv)
{
	int		i;
	int		rc;
//	uint8_t		u8;
//	uint16_t	u16;
//	uint32_t	u32;

	if (argc < 1) return 2;
	i = 0;
//	logging_entry(logger, LOG_LEVEL_DEBUG, "pfm command: %s", argv[i]);
//	logging_entry(logger, LOG_LEVEL_DEBUG, "pfm args=%d", argc);
	if (!strcmp(argv[i], "list")) {
		rc = pnetbuffer_add_uint16(cmd_nb, CMD_LIST_PFMS);
		if (rc != 0) return 2;
	} else if (!strcmp(argv[i], "command")) {
		if (argc < 3) {
			logging_entry(logger,
				LOG_LEVEL_ERROR,
				"pfm command missing required arguments");
			return 2;
		}
		rc = pnetbuffer_add_uint16(cmd_nb, CMD_PFM_COMMAND);
		if (rc != 0) {
			logging_entry(logger,
				LOG_LEVEL_ERROR,
				"failed to build packet content");
			return 2;
		}
		i++;
		rc = pnetbuffer_add_strings(cmd_nb,
					&argv[i],
					1,
					PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
		if (rc != 0) {
			logging_entry(logger,
				LOG_LEVEL_ERROR,
				"failed to build pfm command packet content");
			return 2;
		}
		i++;
		rc = pnetbuffer_add_strings(cmd_nb,
					&argv[i],
					argc-i,
					PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
		if (rc != 0) {
			logging_entry(logger,
				LOG_LEVEL_ERROR,
				"failed to build pfm command packet content");
			return 2;
		}
	} else {
		logging_entry(logger,
			LOG_LEVEL_ERROR,
			"invalid pfm command %s",
			argv[i]);
		return 2;
	}
	return rc;
}
	


int main (int argc, char **argv, char**env)
{
	int		i;
	int		rc;
	uint8_t		u8;
	uint16_t	u16;
	uint16_t	u16_2;
	uint16_t	u16_3;
	uint32_t	u32;
	char		cmd_buf[PACKET_BUFFER_SIZE];
	pnetbuffer_t	*cmd_nb;
	char		res_buf[PACKET_BUFFER_SIZE];
	pnetbuffer_t	*res_nb;
	char		str[256];
	char		*array[2];
//	char		*p;
	compare_t	*c;

	cmd_nb = pnetbuffer_init(cmd_buf, PACKET_BUFFER_SIZE);
	res_nb = pnetbuffer_init(res_buf, PACKET_BUFFER_SIZE);

	if (argc < 2) goto main_exit_usage;
	if (argc > 1) {
		i = 1;
		while (i < argc) {
			if (!strcmp(argv[i], "ping")) {
				rc = pnetbuffer_add_uint16(cmd_nb, CMD_PING);
			} else if (!strcmp(argv[i], "set")) {
				i++;
				if (!strcmp(argv[i], "log-level")) {
					i++;
					u8 = logging_string_to_level(argv[i]);
					rc = pnetbuffer_add_uint16(cmd_nb, CMD_SET_LOG_LEVEL);
					rc = pnetbuffer_add_uint8(cmd_nb, u8);
				}
			} else if (!strcmp(argv[i], "stats")) {
				rc = pnetbuffer_add_uint16(cmd_nb,
							CMD_PACKET_STATS);
			} else if (!strcmp(argv[i], "pfm")) {
				i++;
				rc = arguments_handle_pfm(cmd_nb,
							argc-i,
							&argv[i]);
				if (rc != 0) goto main_exit_usage;
				break; // remainder of args handled
			} else if (!strcmp(argv[i], "filter")) {
				if (argc < 5) goto main_exit_usage;
				i++;
				c = calloc(1,sizeof(compare_t));
				if (!strcmp(argv[i], "add")) {
					i++;
					rc = args_parse_compare_values(c,
							&argv[i],
							argc-i);
					if (rc != 0) {
						free(c);
						goto main_exit_usage;
					}
					rc = pnetbuffer_add_uint16(cmd_nb,
								CMD_FILTER_ADD);
					rc = pnetbuffer_write_compare(cmd_nb, c);
				} else if (!strcmp(argv[i], "del")) {
					i++;
					rc = args_parse_compare_values(c,
							&argv[i],
							argc-i);
					if (rc != 0) {
						free(c);
						goto main_exit_usage;
					}
					rc = pnetbuffer_add_uint16(cmd_nb,
							CMD_FILTER_REMOVE);
					rc = pnetbuffer_write_compare(cmd_nb, c);
				} else {
					logging_entry(logger,
						LOG_LEVEL_ERROR,
						"invalid filter command %s",
						argv[i]);
					goto main_exit_usage;
				}
			} else {
				goto main_exit_usage;
			}
			i++;
		}
	}


//	logging_entry(logger, LOG_LEVEL_DEBUG, "request packet:");
//	pnetbuffer_get_readable(cmd_nb, &p, &u16);
//	dump_memory_hex(p, u16);

	rc = command_and_response(cmd_nb, res_nb);
	if (rc != 0) {
		return 2;
	}

//	logging_entry(logger, LOG_LEVEL_DEBUG, "response packet:");
//	pnetbuffer_get_readable(res_nb, &p, &u16);
//	dump_memory_hex(p, u16);


	//dump_memory_hex(buffer, len);
	rc = pnetbuffer_read_uint16(res_nb, &u16);
	if (!(u16 & CMD_RESPONSE)) {
		logging_entry(logger,
			LOG_LEVEL_ERROR,
			"received packet was not a response");
		return 2;
	}
	u16 &= ~(CMD_RESPONSE);
	switch(u16) {
		case CMD_PING:
			logging_entry(logger,
				LOG_LEVEL_INFO,
				"Ping response received");
			break;
		case CMD_SET_LOG_LEVEL:
			logging_entry(logger,
				LOG_LEVEL_INFO,
				"Log level set");
			break;
		case CMD_PACKET_STATS:
			rc = pnetbuffer_read_uint16(res_nb, &u16);
			for (i=0 ; i<u16 ; i++) {
				rc = pnetbuffer_read_uint16(res_nb, &u16_2);
				rc = pnetbuffer_read_uint32(res_nb, &u32);
				logging_entry(logger,
					LOG_LEVEL_INFO,
					"packet stats ethertype=%04x count=%d",
					u16_2,
					u32);
			}
			break;
		case CMD_FILTER_ADD:
			rc = pnetbuffer_read_uint16(res_nb, &u16);
			logging_entry(logger,
				LOG_LEVEL_INFO,
				"filter addition rc=%d",
				u16);
			break;
		case CMD_FILTER_REMOVE:
			rc = pnetbuffer_read_uint16(res_nb, &u16);
			logging_entry(logger,
				LOG_LEVEL_INFO,
				"filter removal rc=%d",
				u16);
			break;
		case CMD_LIST_PFMS:
			rc = pnetbuffer_read_uint16(res_nb, &u16); // pfm count
			logging_entry(logger,
				LOG_LEVEL_DEBUG,
				"PFM count=%d",
				u16);
			for (u16_2=0 ; u16_2<u16 ; u16_2++) {
				//name len
				rc = pnetbuffer_read_uint8(res_nb, &u8);
				// name
				rc = pnetbuffer_read_data(res_nb, str, u8);
				str[u8] = 0;
				// pfm version
				rc = pnetbuffer_read_uint16(res_nb, &u16_3);
				logging_entry(logger,
					LOG_LEVEL_DEBUG,
					"PFM name=%s version=%d",
					str,
					u16_3);
			}
			break;
		case CMD_PFM_COMMAND:
			rc = pnetbuffer_read_uint16(res_nb, &u16); // return code
			array[0] = str;
			rc = pnetbuffer_read_strings(res_nb,
				array,
				255,
				1,
				&u8,
				PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
			logging_entry(logger,
				LOG_LEVEL_INFO,
				"pfm command response rc=%d(%s)",
				u16,
				array[0]);
			if (u16 != 0) {
				break;
			}
			while (1) {
				rc = pnetbuffer_read_strings(res_nb,
					array,
					255,
					1,
					&u8,
					PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
				if (rc != 0)
					break;
				printf("%s", array[0]); // show raw outout
			}
			break;
		default:
			break;
	}
	
	return 0;

main_exit_usage:
	usage();
	return 2;
}


